from ._backend_calls import _backend_POST, _backend_PUT, _backend_GET, check_backend_availability
from .functions_projects import (
    create_new_project,
    enable_component,
    get_enabled_components,
    get_list_of_all_project_infos,
    get_mission_orbit,
    get_prepared_system_generator_info,
    get_project_info,
    get_recent_project,
    get_sys_arch,
    set_mission_orbit,
    set_new_project,
    set_project_name,
    set_sys_arch,
    set_sys_arch_defaults,
    set_sys_generator,
    set_trained_model,
    traverse_and_modify,
    update_sys_generator,
    )
from .functions import (
    load_file,
    )
from .functions_components import (
    get_tags_from_string,
    comp_create,
    export_components,
    get_comp_statistics
    )

# deprecated functions
from ._backend_calls import _backend_put, _backend_get, _backend_post

# renamed functions
from .renamed_functions import get_mission_info, get_all_decisions, translate_tomlsystem_to_backend, title_with_project_selector, load_project


from .functions_projects import get_list_of_all_project_infos, get_recent_project
from ._backend_calls import _backend_GET
from .functions import move_parameter_placeholder
import streamlit as st

def st_project_selector(columns_sizes: tuple[int, int, int, int] = (5, 1, 2, 2), title: str = None):

    projects_simple_list = get_list_of_all_project_infos()

    col01, col02, col03, col04 = st.columns(columns_sizes)
    if title:
        with col01:
            st.subheader(title)

    with col02:
        if st.button("Get recent project"):
            (st.session_state["project_id"],
             st.session_state["project_name"],
             st.session_state["modified"]) = get_recent_project()
    with col03:
        # Let the user select a project from the list of projects in the project DB
        selected_project = st.selectbox("Project",
                                        projects_simple_list,
                                        index=None,
                                        placeholder="Select a project",
                                        format_func=lambda x: f"{x[1]}")
        if selected_project is not None:
            st.session_state["project_id"] = selected_project[0]
            st.session_state["project_name"] = selected_project[1]
    with col04:
        if 'project_id' not in st.session_state or st.session_state['project_id'] is None:
            st.metric(label="   ",
                  value="<-- Please select")
        else:
            st.metric(label=f"Active project ID: {st.session_state['project_id']}",
                      value=f"{st.session_state['project_name']}")

def st_generated_design_selector(columns_sizes: [int, int, int, int] = [5, 1, 2, 2]):
    """A function which adds a generated design selector to the top of the page.
    A generated design is the stored as part of the project in the backend.

    Suitable to be used in conjunction with st_project_selector().
    """
    # check if the project is set and exit without doing anything if not
    if 'project_id' not in st.session_state or st.session_state['project_id'] is None:
        return

    col01, col02, col03, col04 = st.columns(columns_sizes)

    # get the list of system generators for the current project
    status_code, system_generators = _backend_GET(f"/projects/{st.session_state['project_id']}/sysgen")

    # if successful display a selector for the system generators based on their IDs and names
    with col03:
        if status_code == 200:
            selected_sysgen = st.selectbox("System Generator",
                                                [(sysgen["id"], sysgen["description"])
                                                 for sysgen in system_generators],
                                                index=None,
                                                placeholder="Select a system generator",
                                                format_func=lambda x: f"{x[1]}")

            if selected_sysgen is not None:
                st.session_state["sysgen_id"] = int(selected_sysgen[0])
                st.session_state["sysgen_description"] = selected_sysgen[1]
        else:
            st.error(f"Error while loading system generators: {system_generators}")

    # After selecting a system generator, extract and present the generated designs from system_generators with the ID system_generators_id
    if 'sysgen_id' in st.session_state:
        status_code, system_generator = _backend_GET(f"/projects/{st.session_state['project_id']}/sysgen/{st.session_state['sysgen_id']}/")
        if status_code == 200:
            st.session_state["system_generator"] = system_generator
            st.session_state["generated_designs"] = system_generator.get("generated_designs")
        else:
            st.error(f"Error while loading generated designs: {system_generator}")

    # display the generated designs (comp_list_uids) in a selectbox
    with col04:
        # Only show design selector and warnings if a system generator has been selected
        if 'sysgen_id' in st.session_state:
            if 'generated_designs' in st.session_state and st.session_state["generated_designs"]:
                generated_designs = st.selectbox("Generated Design",
                                                 [(design["id"], design["comp_list_uids"])
                                                  for design in st.session_state["generated_designs"]],
                                                 index=None,
                                                 placeholder="Select a generated design",
                                                 format_func=lambda x: f"Design #{x[0]}, Components: {x[1]}")

                if generated_designs is not None:
                    st.session_state["generated_design_id"] = generated_designs[0]
                    st.session_state["generated_design_comp_list_uids"] = generated_designs[1]
            else:
                st.warning("No generated designs available for the selected system generator.")
    
    # display the selected system generator and generated design
    if 'sysgen_description' in st.session_state and 'generated_design_comp_list_uids' in st.session_state:
        st.metric(label=f"Selected System Generator (ID: {st.session_state['sysgen_id']})",
                  value=st.session_state["sysgen_description"])
        st.metric(label=f"Selected Generated Design (ID: {st.session_state['generated_design_id']})",
                  value=str(st.session_state["generated_design_comp_list_uids"]))
    elif 'sysgen_description' in st.session_state:
        st.metric(label="Selected System Generator",
                  value=st.session_state["sysgen_description"])


def st_load_components(uid_list: list[str] = None) -> None:
    """Loads the defined UIDs from the backend into the session state under session_state['db']['components'][{uid}]
    If no list of UIDs is given, check the UIDs in generated_design_comp_list_uids and load them instead.

    Displays a cute spinner while loading.
    """

    # exit if no data is given to act upon (selection has not been made previously)
    if 'generated_design_comp_list_uids' not in st.session_state and uid_list is None:
        return

    # create sessions state db if not exists
    if 'db' not in st.session_state:
        st.session_state['db'] = {}
    if 'components' not in st.session_state['db']:
        st.session_state['db']['components'] = {}

    # show infobox with loading message
    with st.spinner("Loading components..."):
        if uid_list is None and 'generated_design_comp_list_uids' in st.session_state:
                for component_uid in st.session_state['generated_design_comp_list_uids']:
                    status_code, component = _backend_GET(f"/v2/components/{component_uid}")
                    if status_code == 200:
                        st.session_state['db']['components'][component_uid] = component
                    else:
                        st.error(f"Error loading component {component_uid}: {component}")

        else:
            # Load each component by UID
            for component_uid in uid_list:
                status_code, component = _backend_GET(f"/v2/components/{component_uid}")
                if status_code == 200:
                    st.session_state['db']['components'][component_uid] = component
                else:
                    st.error(f"Error loading component {component_uid}: {component}")

    for component_uid, component in st.session_state['db']['components'].items():
        st.session_state['db']['components'][component_uid] = move_parameter_placeholder(component)


import ast
import json
from pathlib import Path
import toml
import pandas as pd

from ._backend_calls import _backend_POST, _backend_PUT, _backend_GET
from .functions import load_file

def create_new_project(project_info: dict):
    """Create a new project in the backend.
    Params:
        project_info(dict):
    Returns:
        dict:
    """
    status_code, message = _backend_POST(
        endpoint=f"/v2/projects/", data=project_info)
    if status_code == 200:
        return message
    else:
        print(f"Error creating project: {message}")
        return {"error": message}


def get_list_of_all_project_infos() -> list[tuple[int, str, str]]:
    # check for refactoring
    """A function to connect to the backend and get a list of all projects
    in the current project data base.

    Params:
        None

    Returns:
        tuple[int, str, str]: A list of tuples containing project ID, project name, and last modified date.
    """

    status_code, projects = _backend_GET(endpoint=f"/v2/projects/")
    project_ids: list[tuple[int, str, str]] = []
    if status_code == 200:
        for project in projects:
            project_ids.append((project["id"], project["project_name"], project["modified"]))
        return project_ids
    else:
        print(f"Error fetching projects: {projects}")
        return []

def get_project_info(project_id: int) -> dict:
    status_code, project_info = _backend_GET(endpoint=f"/v2/projects/{project_id}/")
    return project_info


def set_project_name(project_id: int, project_name: str) -> dict:
    """A function to set the project name of a specific project, identified by its ID.

    Params:
        project_id (int): The project Identification Number
        project_name(str): Name of the project/mission, eg OPS-Sat
    Returns:
        dict: json response from backend
    """

    data = {
        "project_id": project_id,
        "project_name": project_name,
    }

    # status, msg = _backend_put(endpoint=f"/v2/projects/{project_id}", data=data)
    # return (status, msg)
    status_code, msg = _backend_PUT(endpoint=f"/v2/projects/{project_id}/", data=data)
    if status_code != 200:
        print(f"Error setting project name: {msg}")
        return {"error": msg}
    return msg

def get_recent_project() -> tuple[int, str, str]:
    status_code, response = _backend_GET(endpoint=f"/v2/projects/recent/")
    if status_code != 200:
        print(f"Error fetching recent project: {response}")
        return (None, "No recent project found.")
    return (response["id"], response["project_name"], response["modified"])


def get_mission_orbit(project_id: int) -> dict:
    # refactored
    """A function to get the mission orbit info for a specific project based on its ID.
    Params:
        project_id (int): The project Identification Number

    Returns:
        dict: Keplerian elements to define the orbit
    """

    status_code, response = _backend_GET(endpoint=f"/v2/projects/{project_id}/mission/")
    if status_code != 200:
        print(f"Error fetching mission/orbit info: {response}")
        return {"error": response}
    return response["orbit"]


def set_mission_orbit(project_id: int, orbit_info: dict) -> dict:
    # refactored
    """A function to set the orbit in the project database. Will create an empty mission parent if it does not exist.
    Params:
        project_id (int): The project Identification Number
        orbit_info(dict): Information about the satellite's orbit
    Returns:
        dict: 
     """
    # check if project exists
    status_code, project_info = _backend_GET(endpoint=f"/v2/projects/{project_id}/")
    if status_code != 200:
        print(f"Error fetching project info: {project_info}")
        return {status_code: project_info}
    status_code, mission_info = _backend_GET(
        endpoint=f"/v2/projects/{project_id}/mission/")
    if status_code == 404:
        print(f"No mission info found for project {project_id}. Creating new mission info.")
        mission_info = {
            "mission_name": f"Mission {project_info['project_name']}",
            "orbit": orbit_info,
        }
        status_code, response = _backend_POST(endpoint=f"/v2/projects/{project_id}/mission/", data=mission_info)
        if status_code != 201:
            print(f"Error creating mission info: {response}")
            return {status_code: response}
        return response
    elif status_code == 200:
        # update possible mission_info["orbit"] with new orbit_info
        print(f"Mission info found for project {project_id}. Updating orbit info.")
        mission_info["orbit"].update(orbit_info)
        status_code, response = _backend_PUT(endpoint=f"/v2/projects/{project_id}/mission/", data=mission_info)
        if status_code != 200:
            print(f"Error updating mission info: {response}")
            return {status_code: response}
        return response
    else:
        print(f"Error fetching mission info: {mission_info}")
        return {status_code: mission_info}


def get_enabled_components(nested_dict, enabled_components=None) -> list:
    # check for refactoring
    """
    Recursively traverses a nested dictionary to find and return a list of components that are enabled.

    Parameters:
    nested_dict (dict): The nested dictionary to traverse.
    enabled_components (list, optional): A list to store the names of the enabled components.
                                         Defaults to None, in which case a new list is created.

    Returns:
    list: A list of the names of the enabled components.
    """

    if enabled_components is None:
        enabled_components = []

    for key, value in nested_dict.items():
        if isinstance(value, dict):
            if value.get('Enabled') == True:
                enabled_components.append(key)
                pass
            get_enabled_components(value, enabled_components)

    return enabled_components




def traverse_and_modify(d: dict, sys_config_enabled: dict):
    # check for refactoring
    for key, value in d.items():
        if isinstance(value, dict):
            component = key
            traverse_and_modify(d=value, sys_config_enabled=sys_config_enabled)
        else:
            # This block executes only if `value` is not a dict, i.e., at the deepest level
            # if d["Enabled"] is True or d["Enable"] is True or :
            try:
                if d["Enabled"] is True:
                    sys_config_enabled.update(enable_component(
                        component_name=key, data=load_file(Path("data/backend_sys_default.json"))))
            except KeyError as e:
                print("Error: ", e)


def enable_component(component_name: str, data: dict) -> dict:
    # check for refactoring
    """Recursively search for a component in the nested dictionary from the backend
    and set 'enabled' to True if the feature name matches any key or value (case-insensitive).

    Parameters:
        data (dict): The nested dictionary to search within.
        feature_name (str): The feature name to search for, case-insensitively.

    Returns:
        dict: edited dict
    """

    for key, value in data.items():
        #  print(key, value)
        if isinstance(value, dict):
            enable_component(component_name, value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    enable_component(component_name, item)
        if key.lower() == component_name.lower() or str(value).lower() == component_name.lower():
            data['enabled'] = True
    return data


def set_sys_arch(project_id: int, sys_arch: dict) -> dict:
    # refactored
    """A function to set the system architecture in the project database. Updates the existing architecture if it exists.
    Params:
        project_id (int): The project Identification Number
        sys_arch(dict): Information about the satelllite's modules, which form the system architecture
    Returns:
        dict: Backend version of the system architecture or the status_code and error message.
     """

    sys_arch_example = {
        "system_type": "string",
        "satellite_class": "string",
        "OAP_W": 0,
        "requirements": {},
        "aodcs": {
          "enabled": False,
          "amount_of_entities": 0,
          "description": "string",
          "eFunctions": [],
          "type": "string",
          "components": [],
          "depending_parameters": [],
          "cmg": {
            "enabled": False,
            "amount_of_entities": 0,
            "description": "string",
            "eFunctions": [],
            "requirements": {}
          },
          "magnetorquers": {
            "enabled": False,
            "amount_of_entities": 0,
            "description": "string",
            "eFunctions": [],
            "requirements": {}
          },
        },
        "tcs": {
          "enabled": False,
          "amount_of_entities": 0,
          "description": "string",
          "eFunctions": [],
          "requirements": {},
          "coatings": {
            "enabled": False,
            "amount_of_entities": 0,
            "description": "string",
            "eFunctions": [],
            "requirements": {}
          },
          "heaters": {
            "enabled": False,
            "amount_of_entities": 0,
            "description": "string",
            "eFunctions": [],
            "requirements": {}
          }
        }
    }

    status_code, response = _backend_POST(
        endpoint=f"/v2/projects/{project_id}/system/", data=sys_arch)
    if status_code == 201:
        return response
    if status_code == 400:
        status_code, response = _backend_PUT(
            endpoint=f"/v2/projects/{project_id}/system/", data=sys_arch)
        if status_code == 200:
            return response

    return {status_code: response}


def get_sys_arch(project_id: int) -> dict:
    """A function to get the system configuration info for a specific project based on its ID.
    Params:
        project_id (int): The project Identification Number

    Returns:
        dict: System information
    """

    status_code, response = _backend_GET(
        endpoint=f"/v2/projects/{project_id}/system/")
    return response

#  set_comp_list(project: int, not_yet_defined: dict)
# For each selected system generator, we need a place to store the corresponding found comp lists
# Every generator can produce n sets of components


def set_sys_generator(project_id: int, sys_gen_info: dict) -> dict:
    response = _backend_POST(
        endpoint=f"/v2/projects/{project_id}/sysgen/", data=sys_gen_info)
    return response


def update_sys_generator(project_id: int, sys_gen_info: dict) -> dict:
    response = _backend_PUT(
        endpoint=f"/v2/projects/{project_id}/sysgen/", data=sys_gen_info)
    return response


def set_trained_model(project_id: int, model_info: dict) -> dict:
    """A function to upload a trained AI model.
    Params:
        project_id (int): The project Identification Number
        model (ASKYOUNES): info about model

    Returns:
        dict: Upload server response
    """

    sta, msg = _backend_POST(endpoint=f"/v2/projects/{project_id}/sysgen/",
                             data=model_info)

    return msg


def get_prepared_system_generator_info(project_id: int) -> list:
    """A function to get all prepared system generators.
    Params:
        project_id (int): The project Identification Number

    Returns:
        list: list of dictionaries with infos for the prepared system generators
    """

    status_code, response = _backend_GET(
        endpoint=f"/v2/projects/{project_id}/sysgen/")
    return response


def set_sys_arch_defaults(system: dict, project_id: int) -> dict:
    """Translate a system configuration from a TOML file dictionary previously handled by load_file() to a format that can be uploaded to the backend.

    Parameters:
    system (dict): The system configuration in TOML format.

    Returns:
    dict: The system configuration in a format that can be uploaded to the backend.
    """

    sys_config_enabled = {}
    # Loading System Configuration
    for key, value in system.items():
        if isinstance(value, dict):
            traverse_and_modify(d=value, sys_config_enabled=sys_config_enabled)

    response = set_sys_arch(project_id=project_id, sys_arch=sys_config_enabled)
    pass
    return response


def set_new_project(project_info: dict) -> dict:
    """Create a new project in the backend.
    Params:
        project_info(dict):
    Returns:
        dict:
    """
    status_code, response = _backend_POST(
        endpoint="/v2/projects/", data=project_info)
    if status_code == 201:
        return response
    return {status_code: response}


# def get_trained_model(project_id: int) -> mode: ASKYOUNES
#     pass


# def set_train_log(project_id: int, logs: ASKYOUNES) -> DB_response: ASKALEX
#     pass


# def get_train_logs(project_id: int) -> logs: ASKYOUNE   pass




import ast
import json
from pathlib import Path
import toml
import pandas as pd

from ._backend_calls import _backend_POST, _backend_PUT, _backend_GET



def get_tags_from_string(str_of_tags: str) -> list[str]:
    """
    example_tags_str = "['satellite', 'power', 'solar-panels']"
    check https://gitlab.com/ai4ce/ai4ce-03-backend/-/blob/main/backend/v2/schemas/components/_uid_maps.py for valid tags
    """
    # The string representation of a list of tags

    # Convert the string to an actual list using ast.literal_eval
    tags: list = ast.literal_eval(str_of_tags)
    return tags


def comp_create(comp_info: dict) -> dict:
    """Create a new component in the backend.
    Params:
        comp_info(dict): Information about the component
    Returns:
        dict:
    """
    status_code, msg = _backend_POST(
        endpoint=f"/v2/components/", data=comp_info)
    return status_code, msg

def export_components() -> dict:
    """Fetches all decisions from the DecisionDB and returns them as a dictionary.

    Returns:
        dict: A dictionary containing all decisions fetched from the DecisionDB.
    """
    status_code, response = _backend_GET(endpoint="/v2/components/export/")
    return response

def get_comp_statistics() -> dict:

    status_code, response = _backend_GET(endpoint="/v2/streamlit/db/stats/components")
    return response


def split_uid(uid: str) -> tuple[str, str, str, int | str]:
    """Splits a component UID into its constituent parts: type, subtype, and name.

    Args:
        uid (str): The UID of the component in the format 'type-subtype-name'.

    Returns:
        tuple: A tuple containing the type, subtype, and name of the component.
    """
    try:
        system_uid, subsystem_uid, component_class, component_id = uid.split("_", 3)
        try:
            component_id = int(component_id)
        except ValueError:
            pass
        return system_uid, subsystem_uid, component_class, component_id
    except ValueError:
        raise ValueError("Invalid UID format. Expected UID format: <system>_<subsystem>_<component_class>_<component_id>")